use core::fmt::Debug;
use alloc::{boxed::Box, string::{String, ToString}, vec::Vec, vec};
use hashbrown::HashMap;
use spin::{Mutex, MutexGuard};

use crate::utils::as_any_mut;

pub mod ps2;

lazy_static::lazy_static!(
    pub static ref DRIVER_MGR: Mutex<DriverHandler> = Mutex::new(DriverHandler::new());
);

type RegisterFn = fn(
    name: &mut String,
    author: &mut String,
    license: &mut String,
) -> Box<dyn Driver>;

pub trait Driver: Debug + Send{
    fn new() -> impl Driver where Self: Sized;
    fn register(
        name: &mut String,
        author: &mut String,
        license: &mut String,
    ) -> Box<dyn Driver> where Self: Sized {
        *name = alloc::format!("YOUR_DRIVER_NAME");
        *author = alloc::format!("Your Name <your@email.tld>");
        *license = alloc::format!("GPLv3");

        // Box::new(Self::new())
        todo!()
    }
    fn startup(&mut self) -> anyhow::Result<()> {
        Ok(())
    }
    fn shutdown(&mut self) -> anyhow::Result<()> {
        Ok(())
    }

}


#[derive(Debug)]
pub struct DriverHandler {
    drivers: HashMap<String, Box<dyn Driver>>,
}

impl DriverHandler {
    pub fn new() -> Self {
        Self {
            drivers: HashMap::new(),
        }
    }
    
    #[inline]
    pub fn get_instance<'i>() -> MutexGuard<'i, Self> {
        DRIVER_MGR.lock()
    }

    pub fn load_drivers(&mut self) {
        for driver_reg_fn in Self::get_all_driver_entries() {
            let mut name = String::new();
            let mut author = String::new();
            let mut license = String::new();
            let driver = (driver_reg_fn)(&mut name, &mut author, &mut license);
            log::info!("Loaded driver '{name}' ({license})");
            self.drivers.insert(name, driver);
        }
    }
    
    pub fn get_driver<T: Driver + 'static>(&mut self, name: &str) -> Option<&mut T> {
        if let Some(d) = self.drivers.get_mut(&name.to_string()) {
            unsafe {
                // Lets hope for the best
                Some(as_any_mut(d).downcast_mut_unchecked::<T>())
            }
        } else {
            None
        }
    }

    pub fn start_drivers(&mut self) {
        for (name, driver) in &mut self.drivers {
            if let Err(e) = driver.startup() {
                log::error!("Driver '{name}' had an error on startup: {e}");
            }
        }
    }

    pub fn stop_drivers(&mut self) {
        for (name, driver) in &mut self.drivers {
            if let Err(e) = driver.shutdown() {
                log::error!("Driver '{name}' had an error on shutdown: {e}");
            }
        }
    }

    #[inline]
    fn get_all_driver_entries() -> Vec<RegisterFn> {
        vec![
            ps2::Ps2Driver::register,
        ]
    }
}
