use core::ptr::addr_of;
use x86_64::instructions::segmentation::{CS, DS, Segment};
use x86_64::structures::gdt::{Descriptor, DescriptorFlags, GlobalDescriptorTable, SegmentSelector};
use x86_64::VirtAddr;
use x86_64::structures::tss::TaskStateSegment;
use lazy_static::lazy_static;


pub const STACK_SIZE: usize = 4096 * 5;
pub static mut STACK: [u8; STACK_SIZE] = [0; STACK_SIZE];
pub static mut PRIV_TSS_STACK: [u8; STACK_SIZE] = [0; STACK_SIZE];



pub const DOUBLE_FAULT_IST_INDEX: u16 = 0;

lazy_static! {
    static ref TSS: TaskStateSegment = {
        let mut tss = TaskStateSegment::new();
        tss.interrupt_stack_table[DOUBLE_FAULT_IST_INDEX as usize] = {
            
            #[allow(unused_unsafe)]
            let stack_start = VirtAddr::from_ptr(unsafe { addr_of!(STACK) });
            stack_start + STACK_SIZE as u64
        };
        tss.privilege_stack_table[0] = {
            #[allow(unused_unsafe)]
            let stack_start = VirtAddr::from_ptr(unsafe { addr_of!(PRIV_TSS_STACK) });
            stack_start + STACK_SIZE as u64
        };
        tss
    };
}


lazy_static! {
    pub static ref GDT: (GlobalDescriptorTable, Selectors) = {
        let mut gdt = GlobalDescriptorTable::new();
        let kernel_data_flags = DescriptorFlags::USER_SEGMENT | DescriptorFlags::PRESENT | DescriptorFlags::WRITABLE;

        let code_sel = gdt.append(Descriptor::kernel_code_segment());
        let data_sel = gdt.append(Descriptor::UserSegment(kernel_data_flags.bits()));

        let tss_sel = gdt.append(Descriptor::tss_segment(&TSS));
        let user_data_sel = gdt.append(Descriptor::user_data_segment()); // user data segment
        let user_code_sel = gdt.append(Descriptor::user_code_segment()); // user code segment
        (gdt, Selectors {code_sel, data_sel, tss_sel, user_data_sel, user_code_sel })
    };
}

#[allow(dead_code)]
#[derive(Debug)]
pub struct Selectors {
    pub code_sel: SegmentSelector,
    pub data_sel: SegmentSelector,
    pub tss_sel: SegmentSelector,
    pub user_data_sel: SegmentSelector,
    pub user_code_sel: SegmentSelector
}

pub fn init() {
    use x86_64::instructions::tables::load_tss;
    #[allow(unused_unsafe)]
    let stack = unsafe { addr_of!(STACK) as *const _  };
    #[allow(unused_unsafe)]
    let user_stack = unsafe { addr_of!(PRIV_TSS_STACK) as *const _ };
    log::info!("gdt init");
    log::debug!("gdt             = {:?}", &GDT.0 as *const _ );
    log::debug!("tss             = {:?}", &*TSS as *const _ );
    log::debug!("kernel_stack    = {:?}", stack);
    log::debug!("user_stack      = {:?}", user_stack);
    log::debug!("kernel_code_sel = {:?}", GDT.1.code_sel);
    log::debug!("kernel_data_sel = {:?}", GDT.1.data_sel);
    log::debug!("tss_sel         = {:?}", GDT.1.tss_sel);
    log::debug!("user_code_sel   = {:?}", GDT.1.user_code_sel);
    log::debug!("user_code_sel   = {:?}", GDT.1.user_data_sel);


    GDT.0.load();
    unsafe {
        DS::set_reg(GDT.1.data_sel);
        CS::set_reg(GDT.1.code_sel);
        load_tss(GDT.1.tss_sel);
    }
}



