use pic8259::ChainedPics;
use x86_64::structures::idt::InterruptDescriptorTable;


mod handlers;

pub const PIC_1_OFFSET: u8 = 32;
pub const PIC_2_OFFSET: u8 = PIC_1_OFFSET + 8;


lazy_static::lazy_static! {
    static ref IDT: InterruptDescriptorTable = {
        let mut idt = InterruptDescriptorTable::new();

        idt.breakpoint.set_handler_fn(handlers::breakpoint);
        idt.page_fault.set_handler_fn(handlers::page_fault);
        idt[InterruptIndex::Timer.get()]
            .set_handler_fn(handlers::timer::timer_interrupt);

        idt[InterruptIndex::Ps2Keyboard.get()]
            .set_handler_fn(handlers::ps2::ps2_kb_int);


        unsafe {
            idt.double_fault.set_handler_fn(handlers::double_fault)
                .set_stack_index(crate::gdt::DOUBLE_FAULT_IST_INDEX);
        }
        idt
    };
}

pub static PICS: spin::Mutex<ChainedPics> =
        spin::Mutex::new(unsafe { ChainedPics::new(PIC_1_OFFSET, PIC_2_OFFSET) });


/// https://wiki.osdev.org/Interrupts
/// | IRQ  | Description                                              |
/// |------|----------------------------------------------------------|
/// |  0   | Programmable Interrupt Timer Interrupt                   |
/// |  1   | Keyboard Interrupt                                       |
/// |  2   | Cascade (used internally by the two PICs. never raised)  |
/// |  3   | COM2 (if enabled)                                        |
/// |  4   | COM1 (if enabled)                                        |
/// |  5   | LPT2 (if enabled)                                        |
/// |  6   | Floppy Disk                                              |
/// |  7   | LPT1 / Unreliable "spurious" interrupt (usually)         |
/// |  8   | CMOS real-time clock (if enabled)                        |
/// |  9   | Free for peripherals / legacy SCSI / NIC                 |
/// |  10  | Free for peripherals / SCSI / NIC                        |
/// |  11  | Free for peripherals / SCSI / NIC                        |
/// |  12  | PS2 Mouse                                                |
/// |  13  | FPU / Coprocessor / Inter-processor                      |
/// |  14  | Primary ATA Hard Disk                                    |
/// |  15  | Secondary ATA Hard Disk                                  |
#[allow(dead_code)]
#[derive(Debug, Clone, Copy)]
#[repr(u8)]
pub enum InterruptIndex {
    Timer = 0,
    Ps2Keyboard,
    Cascade,
    Com2,
    Com1,
    Lpt2,
    FloppyDisk,
    Lpt1,
    Cmos,
    Ffp0,
    Ffp1,
    Ffp2,
    Ps2Mouse,
    Fpu,
    PrimAtaHdd,
    SecAtaHdd,
}

impl InterruptIndex {
    pub fn get(self) -> u8 {
        (self as u8) + PIC_1_OFFSET
    }
}

pub fn init_idt() {
    log::info!("IDT init");
    IDT.load();
    unsafe { PICS.lock().initialize() };
    x86_64::instructions::interrupts::enable();
}


