#![no_std]
#![no_main]
#![feature(abi_x86_interrupt)]
#![feature(associated_type_defaults)]
#![allow(internal_features)]
#![feature(ptr_internals)]
#![feature(allocator_api)]
#![feature(downcast_unchecked)]

use std::{input::Keyboard, vga::video::Color};

use mem::paging::ActivePageTable;
use multiboot2::{BootInformation, BootInformationHeader};
use pc_keyboard::KeyCode;
use embedded_graphics::prelude::*;
use scheduler::{executor::Executor, Task};
use utils::acpi_shutdown;

extern crate alloc;

#[macro_use]
mod logger;
mod interrupts;
mod gdt;
mod utils;
#[macro_use]
mod events;
mod mem;
mod scheduler;
mod std;
mod drivers;
mod acpi;

fn init(mb2p: *const BootInformationHeader) -> ActivePageTable {
    logger::init(log::LevelFilter::Trace).unwrap();
    gdt::init();
    interrupts::init_idt();
    utils::enable_nxe_bit();
    utils::enable_write_protect_bit();
    std::init();
    let active_table = mem::init(mb2p);
    {
        let mut dm = drivers::DriverHandler::get_instance();
        dm.load_drivers();
        dm.start_drivers();
    }
    active_table
}

fn shutdown() -> ! {
    {
        let mut dm = drivers::DriverHandler::get_instance();
        dm.stop_drivers();
    }
    

    acpi_shutdown()
}

#[no_mangle]
extern "C" fn kmain(mb2p: *const BootInformationHeader) -> ! {
    // let sch = Executor::new();
    let _ = init(mb2p);


    {   
        let mut dm = drivers::DriverHandler::get_instance();
        let driver = dm.get_driver::<drivers::ps2::Ps2Driver>("PS/2");
        if let Some(driver) = driver {
            log::debug!("{:?}", driver.test());
        }
    } 
    
    'mainloop: loop {
        if Keyboard::is_key_down(KeyCode::A) {
           // vc.fill_screen(Color::new(0xFF, 0, 0));
        } else {
            //vc.fill_screen(Color::new(0, 0xFF, 0));
        }

        if Keyboard::are_keys_down(&[KeyCode::LAlt, KeyCode::F4]) {
            break 'mainloop;
        }
    }
    shutdown()
}


#[panic_handler]
fn panic(pi: &core::panic::PanicInfo) -> ! {
    log::error!("{pi}");
    utils::hcf();
}


