use core::alloc::{GlobalAlloc, Layout};

use multiboot2::{BootInformation, BootInformationHeader, MemoryArea, MemoryAreaType, MemoryMapTag, TagTrait};
use talc::{ErrOnOom, Span, Talc, Talck};
use crate::mem::paging::entry::EntryFlags;

use self::heap::{HEAP_SIZE, HEAP_START};
pub use self::paging::remap_the_kernel;

pub mod heap;
pub mod area_frame_alloc;
pub mod paging;

//#[global_allocator]
//static GLOBAL_ALLOCATOR: heap::BumpAllocator = heap::BumpAllocator::new(HEAP_START, HEAP_START + HEAP_SIZE);

#[global_allocator]
static GLOBAL_ALLOCATOR: Talck<spin::Mutex<()>, ErrOnOom> = Talc::new(ErrOnOom).lock();


pub struct DummyAlloc;
unsafe impl GlobalAlloc for DummyAlloc {
    unsafe fn alloc(&self, _: Layout) -> *mut u8 { 
        core::ptr::null_mut::<u8>()
    }
    unsafe fn dealloc(&self, _ptr: *mut u8, _: Layout) {}
}

pub const PAGE_SIZE: usize = 4096;

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
pub struct Frame {
    number: usize,
}


impl Frame {
    pub fn containing_address(address: usize) -> Frame {
        Frame{ number: address / PAGE_SIZE }
    }
    fn start_address(&self) -> paging::PhysicalAddress {
        self.number * PAGE_SIZE
    }
    fn clone(&self) -> Frame {
        Frame { number: self.number }
    }
    fn range_inclusive(start: Frame, end: Frame) -> FrameIter {
        FrameIter {
            start,
            end,
        }
    }
}

struct FrameIter {
    start: Frame,
    end: Frame,
}

impl Iterator for FrameIter {
    type Item = Frame;

    fn next(&mut self) -> Option<Frame> {
        if self.start <= self.end {
            let frame = self.start.clone();
            self.start.number += 1;
            Some(frame)
        } else {
            None
        }
    }
 }


#[allow(dead_code)]
pub trait FrameAllocator {
    fn allocate_frame(&mut self) -> Option<Frame>;
    fn deallocate_frame(&mut self, frame: Frame);
}


#[derive(Clone, Debug)]
pub struct MemoryAreaIter {
    current_area: *const MemoryArea,
    last_area: *const MemoryArea,
    entry_size: u32,
}

impl Iterator for MemoryAreaIter {
    type Item = &'static MemoryArea;
    fn next(&mut self) -> Option<&'static MemoryArea> {
        if self.current_area > self.last_area {
            None
        } else {
            let area = unsafe{&*self.current_area};
            self.current_area = ((self.current_area as u32) + self.entry_size)
                as *const MemoryArea;
            if area.typ() == MemoryAreaType::Available {
                Some(area)
            } else {self.next()}
        }
    }
}

impl MemoryAreaIter {
    pub fn new(mmap_tag: &MemoryMapTag) -> Self {
        let self_ptr = mmap_tag as *const MemoryMapTag;
        let start_area = (mmap_tag.memory_areas().first().unwrap()) as *const MemoryArea;
        MemoryAreaIter {
            current_area: start_area,
            last_area: ((self_ptr as *const () as usize) + mmap_tag.size() - mmap_tag.entry_size() as usize) as *const MemoryArea,
            entry_size: mmap_tag.entry_size(),
        }
    }

}


#[derive(Debug)]
pub struct MemInfo<'a>{
    boot_info: BootInformation<'a>,
    mem_area_iter: MemoryAreaIter,
    pub kernel_start: usize,
    pub kernel_end: usize,
    pub mb_start: usize,
    pub mb_end: usize,
    pub vga_vide_start: usize,
    pub vga_vide_size: usize,
}

impl MemInfo<'_> {
    pub fn load(mb_ptr: *const BootInformationHeader) -> Self {
        let boot_info = unsafe {
            multiboot2::BootInformation::load(mb_ptr).unwrap()
        };
        let mmap_tag = boot_info.memory_map_tag().unwrap();

        // log::debug!("Memory areas:");
        // for area in mmap_tag.memory_areas() {
        //    log::debug!(" start: 0x{:x}, sz: 0x{:x}", area.start_address(), area.size());
        // }

        // log::debug!("kernel sections:");
        // for section in boot_info.elf_sections().unwrap() {
        //     log::debug!("{}: start: 0x{:x}, sz: 0x{:x}, flags: 0b{:b}", section.name().unwrap_or("NONE"),
        //         section.start_address(), section.size(), section.flags());
        // }
        let elf_secs =  boot_info.elf_sections().unwrap(); 
        let kernel_start = elf_secs.clone().filter(|s| s.is_allocated()).map(|s| s.start_address()).min().unwrap() as usize;
        let kernel_end = elf_secs.clone().filter(|s| s.is_allocated()).map(|s| s.end_address()).min().unwrap() as usize;
        let mb_start = boot_info.start_address();
        let mb_end = boot_info.end_address();
        let vv_info = boot_info.framebuffer_tag().unwrap().unwrap();
        let vv_start = vv_info.address();
        let vv_size = vv_info.pitch() * vv_info.height();

        //log::debug!("Kernel: start: 0x{:x} sz: 0x{:x}", mi.kernel_start, mi.kernel_end - mi.kernel_start);
        //log::debug!("Multiboot: start: 0x{:x} sz: 0x{:x}", mi.mb_start, mi.mb_end - mi.mb_start);
        
        Self {
            boot_info: unsafe { multiboot2::BootInformation::load(mb_ptr).unwrap()},
            mem_area_iter: MemoryAreaIter::new(mmap_tag),
            kernel_start,
            kernel_end,
            mb_start,
            mb_end,
            vga_vide_start: vv_start as usize,
            vga_vide_size: vv_size as usize
        }
    }
}

pub fn init(mb_ptr: *const BootInformationHeader) -> paging::ActivePageTable {
    once::assert_has_not_been_called!("mem::init must be called only once");
    let mem_info = MemInfo::load(mb_ptr);
    let mut frame_alloc = area_frame_alloc::AreaFrameAllocator::new(&mem_info);
    
    let mut active_table = remap_the_kernel(&mut frame_alloc, &mem_info);

    use self::paging::Page;

    let heap_start_page = Page::containing_address(HEAP_START);
    let heap_end_page = Page::containing_address(HEAP_START + HEAP_SIZE-1);

    for page in Page::range_inclusive(heap_start_page, heap_end_page) {
        active_table.map(page, EntryFlags::WRITABLE, &mut frame_alloc);
    }
    unsafe {
        GLOBAL_ALLOCATOR.lock().claim(Span::from_base_size(HEAP_START as *mut u8, HEAP_SIZE)).unwrap();
    }
    active_table
}
