use core::{any::Any, arch::asm};

use x86::msr::{rdmsr, wrmsr, IA32_EFER};
use x86_64::{instructions::{self, port::PortWriteOnly}, registers::{control::{Cr0, Cr0Flags}, segmentation::{Segment, DS}}, PrivilegeLevel};

use crate::mem::paging::{ActivePageTable, VirtualAddress};



pub fn hcf() -> ! {
    loop {
        instructions::hlt();
        core::hint::spin_loop();
    }
}

pub fn enable_nxe_bit() {

    let nxe_bit = 1 << 11;
    unsafe {
        let efer = rdmsr(IA32_EFER);
        wrmsr(IA32_EFER, efer | nxe_bit);
    }
}
pub fn enable_write_protect_bit() {

    unsafe { Cr0::write(Cr0::read() | Cr0Flags::WRITE_PROTECT) };
}


pub unsafe fn set_usermode_segments() -> (u16, u16) {
    // set ds and tss, return cs and ds
    let (mut cs, mut ds) = (crate::gdt::GDT.1.user_code_sel, crate::gdt::GDT.1.user_data_sel);
    cs.0 |= PrivilegeLevel::Ring3 as u16;
    ds.0 |= PrivilegeLevel::Ring3 as u16;
    DS::set_reg(ds);
    (cs.0, ds.0)
}


#[no_mangle]
pub unsafe fn jmp_to_usermode(apt: ActivePageTable, code: VirtualAddress, stack_end: VirtualAddress) {
    let (cs_idx, ds_idx) = set_usermode_segments();
    x86_64::instructions::tlb::flush_all(); // flush the TLB after address-space switch
    let code_addr = apt.translate(code).expect("Invalid virt addr");
    let sa_addr = apt.translate(stack_end).expect("Invalid virt addr");

    asm!("push {ds:x}",   // stack segment
        "push {sa}",   // rsp 
        "push 0x200", // rflags (only interrupt bit set)
        "push {cs:x}",   // code segment
        "push {code}",   // ret to virtual addr
        "iretq",
        code = in(reg) code_addr, sa = in(reg) sa_addr, cs = in(reg) cs_idx, ds = in(reg) ds_idx);
}

pub fn acpi_shutdown() -> ! {
    let mut port = x86_64::instructions::port::Port::new(0x0604);
    unsafe {
        port.write(0x2000_u16);
    }
    loop {hcf()}
}

pub fn as_any_mut<T: Any>(i: &mut T) -> &mut dyn Any {
    i as &mut dyn Any
}
